# -*- coding: utf-8 -*-
"""
Created on Tue Jun 05 18:37:23 2012

@author: Serra
"""

#import python modules
import numpy as np
import time

#import compiled fortran library
import projlib
print projlib.__doc__

import sys
import os


import atomwrite
#whether or not to use the 3d visualization
UseVisual = False
if UseVisual:
    #import custom visualization library
    import atomvis


#========== GLOBAL VARIABLES ==========
#set the random number seed; useful for debugging
np.random.seed(342324)


#NOTE:
#everything below assumes unit atomic masses,
#such that forces = accelerations.



def LineSearch(Pos, Dir, dx, EFracTol,
               Accel = 1.5, MaxInc = 10., MaxIter = 10000):
    """Performs a line search along direction Dir.
Input:
    Pos: starting positions, (N,3) array
    Dir: (N,3) array of gradient direction
    dx: initial step amount
    EFracTol: fractional energy tolerance
    Accel: acceleration factor
    MaxInc: the maximum increase in energy for bracketing
    MaxIter: maximum number of iteration steps
Output:
    PEnergy: value of potential energy at minimum along Dir
    Pos: minimum energy (N,3) position array along Dir
"""
    #identify global variables
    global M, L, rw, rlj, rc
    
    #start the iteration counter
    Iter = 0

    #find the normalized direction    
    NormDir = np.clip(Dir, -1.e100, 1.e100)
    NormDir = NormDir / np.sqrt(np.sum(NormDir * NormDir))

    #take the first two steps and compute energies    
    Dists = [0., dx]
    PEs = [projlib.calcenergy(Pos + NormDir * x, AA, M, L, rc, rlj, rw,sigma) for x in Dists]
    
    #if the second point is not downhill in energy, back
    #off and take a shorter step until we find one
    while PEs[1] > PEs[0]:
        Iter += 1
        dx = dx * 0.5
        Dists[1] = dx
        PEs[1] = projlib.calcenergy(Pos + NormDir * dx, AA, M, L, rc, rlj, rw, sigma)
        
    #find a third point
    Dists = Dists + [2. * dx]
    PEs = PEs + [projlib.calcenergy(Pos + NormDir * 2. * dx, AA, M, L, rc, rlj, rw, sigma)]
    
    #keep stepping forward until the third point is higher
    #in energy; then we have bracketed a minimum
    while PEs[2] < PEs[1]:
        Iter += 1
            
        #find a fourth point and evaluate energy
        Dists = Dists + [Dists[-1] + dx]
        PEs = PEs + [projlib.calcenergy(Pos + NormDir * Dists[-1], AA, M, L, rc, rlj, rw, sigma)]

        #check if we increased too much in energy; if so, back off
        if (PEs[3] - PEs[0]) > MaxInc * (PEs[0] - PEs[2]):
            PEs = PEs[:3]
            Dists = Dists[:3]
            dx = dx * 0.5
        else:
            #shift all of the points over
            PEs = PEs[-3:]
            Dists = Dists[-3:]
            dx = dx * Accel
            
    #we've bracketed a minimum; now we want to find it to high
    #accuracy
    OldPE3 = 1.e300
    while True:
        Iter += 1
        if Iter > MaxIter:
            print "Warning: maximum number of iterations reached in line search."
            break
            
        #store distances for ease of code-reading
        d0, d1, d2 = Dists
        PE0, PE1, PE2 = PEs

        #use a parobolic approximation to estimate the location
        #of the minimum
        d10 = d0 - d1
        d12 = d2 - d1
        Num = d12*d12*(PE0-PE1) - d10*d10*(PE2-PE1)
        Dem = d12*(PE0-PE1) - d10*(PE2-PE1)
        if Dem == 0:
            #parabolic extrapolation won't work; set new dist = 0 
            d3 = 0
        else:
            #location of parabolic minimum
            d3 = d1 + 0.5 * Num / Dem
            
        #compute the new potential energy
        PE3 = projlib.calcenergy(Pos + NormDir * d3, AA, M, L, rc, rlj, rw, sigma)
        
        #sometimes the parabolic approximation can fail;
        #check if d3 is out of range < d0 or > d2 or the new energy is higher
        if d3 < d0 or d3 > d2 or PE3 > PE0 or PE3 > PE1 or PE3 > PE2:
            #instead, just compute the new distance by bisecting two
            #of the existing points along the line search
            if abs(d2 - d1) > abs(d0 - d1):
                d3 = 0.5 * (d2 + d1)
            else:
                d3 = 0.5 * (d0 + d1)
            PE3 = projlib.calcenergy(Pos + NormDir * d3, AA, M, L, rc, rlj, rw, sigma)
            
        #decide which three points to keep; we want to keep
        #the three that are closest to the minimum
        if d3 < d1:
            if PE3 < PE1:
                #get rid of point 2
                Dists, PEs = [d0, d3, d1], [PE0, PE3, PE1]
            else:
                #get rid of point 0
                Dists, PEs = [d3, d1, d2], [PE3, PE1, PE2]
        else:
            if PE3 < PE1:
                #get rid of point 0
                Dists, PEs = [d1, d3, d2], [PE1, PE3, PE2]
            else:
                #get rid of point 2
                Dists, PEs = [d0, d1, d3], [PE0, PE1, PE3]
                
        #check how much we've changed
        if abs(OldPE3 - PE3) < EFracTol * abs(PE3):
            #the fractional change is less than the tolerance,
            #so we are done and can exit the loop
            break
        OldPE3 = PE3

    #return the position array at the minimum (point 1)        
    PosMin = Pos + NormDir * Dists[1]
    PEMin = PEs[1]

    #if using visualization, update the display
    if UseVisual:
        if atomvis.Initialized:
            #update the positions
            atomvis.Update(PosMin)
        else:
            #initialize the visualization window
            atomvis.Init(PosMin)
        
    return PEMin, PosMin

        
def ConjugateGradient(Pos, dx, EFracTolLS, EFracTolCG):
    """Performs a conjugate gradient search.
Input:
    Pos: starting positions, (N,3) array
    dx: initial step amount
    EFracTolLS: fractional energy tolerance for line search
    EFracTolCG: fractional energy tolerance for conjugate gradient
Output:
    PEnergy: value of potential energy at minimum
    Pos: minimum energy (N,3) position array 
"""
    #identify global variables
    global M, L, Cut
    #initial search direction
    Forces = np.zeros_like(Pos)
    PE, Forces = projlib.calcenergyforces(Pos, AA, M, L, rc, rlj, rw, sigma, Forces)
    Dir = Forces
    OldPE = 1.e300
    #iterative line searches
    while abs(PE - OldPE) > EFracTolCG * abs(PE):
        OldPE = PE
        PE, Pos = LineSearch(Pos, Dir, dx, EFracTolLS)
        OldForces = Forces.copy()
        PE, Forces = projlib.calcenergyforces(Pos, AA, M, L, rc, rlj, rw, sigma, Forces)
        Gamma = np.sum((Forces - OldForces) * Forces) / np.sum(OldForces * OldForces)
        Dir = Forces + Gamma *  Dir

    Pos[:,0]=Pos[:,0]-25
    Pos[:,1]=Pos[:,1]-5
    return PE, Pos


def InitPositions(N, L):
    """Returns an array of initial positions of each atom,
placed on a cubic lattice for convenience.
Input:
    N: number of atoms
    L: box length
Output:
    Pos: (N,3) array of positions
"""
    #make the position array
    Pos = np.zeros((N,3), float)
    #compute integer grid # of locations for cubic lattice
    NLat = int(N**(1./3.) + 1.)
    LatSpac = L / NLat
    #make an array of lattice sites
    r = LatSpac * np.arange(NLat, dtype=float) - 0.5*L
    #loop through x, y, z positions in lattice until done
    #for every atom in the system
    i = 0
    for x in r:
        for y in r:
            for z in r:
                Pos[i] = np.array([x,y,z], float)
                #add a random offset to help initial minimization
                Offset = 0.1 * LatSpac * (np.random.rand(3) - 0.5)
                Pos[i] = Pos[i] + Offset
                i += 1
                #if done placing atoms, return
                if i >= N:
                    return Pos
             
    return Pos


def RescaleVelocities(Vel, T):
    """Rescales velocities in the system to the target temperature.
Input:
    Vel: (N,3) array of atomic velocities
    T: target temperature
Output:
    Vel: same as above
"""
    #recenter to zero net momentum (assuming all masses same)
    Vel = Vel - Vel.mean(axis=0)
    #find the total kinetic energy
    KE = 0.5 * np.sum(Vel * Vel)
    #find velocity scale factor from ratios of kinetic energy
    VScale = np.sqrt(1.5 * len(Vel) * T / KE)
    Vel = Vel * VScale
    return Vel  


def InitVelocities(N, T):
    """Returns an initial random velocity set.
Input:
    N: number of atoms
    T: target temperature
Output:
    Vel: (N,3) array of atomic velocities
"""
    Vel = np.random.rand(N, 3)
    Vel = RescaleVelocities(Vel, T)
    return Vel


def InitAccel(Pos):
    """Returns the initial acceleration array.
Input:
    Pos: (N,3) array of atomic positions
Output:
    Accel: (N,3) array of acceleration vectors
"""
    #global variables
    global M, L, Cut
    #get the acceleration from the forces
    Forces = np.zeros_like(Pos)
    PEnergy, Accel = projlib.calcenergyforces(Pos, AA, M, L, rc, rlj, rw, sigma, Forces)
    return Accel


def InstTemp(Vel):
    """Returns the instantaneous temperature.
Input:
    Vel: (N,3) array of atomic velocities
Output:
    Tinst: float
"""
    return np.sum(Vel * Vel) / (3. * len(Vel))


#My Code


M = 15
N= 15
#L = (N/0.03)**(1./3) 
L=37.5
P=np.zeros(shape=(N,N)) #matrix to keep track of whether within minimum distance
#P=np.zeros(shape=(3,3)) #matrix to keep track of whether within minimum distance
#P[1,1]+=1
#print P
Pnorm = [] #normalized probability
PErun = [] #check equilibration
#Different configurations used
#AA=np.ones(N)
#AA= np.array([1,2,2,1,1,2,2,2,2,2,1,1,2,2,1]) #Hydrophobic configuration
#AA= np.array([1,3,3,3,3,1,1,1,1,1,4,4,4,4,1]) #Charge configuration
AA= np.array([1,3,1,1,2,2,1,2,2,4,1,1,3,2,4]) #Model all AAs configuration
#AA= np.array([3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4])
#tested if screened coulomb was working with 2 peptides one pos and one neg

Names =[]
k=0
"""while k<M:
    if AA[k] == 1:
        Names.append('Phile')
    if AA [k] ==2:
        Names.append('Phob')
    if AA[k] ==3:
        Names.append('Pos')
    if AA[k]==4:
        Names.append('Neg')
    k+=1  """
while k<M:
    if AA[k] == 1:
        Names.append('H')
    if AA [k] ==2:
        Names.append('C')
    if AA[k] ==3:
        Names.append('P')
    if AA[k]==4:
        Names.append('N')
    k+=1  
#print Names    

f = file("Proj_PEout.txt", "w")
f.write('PE\tTime\n')
g= file("Proj_Eout.txt",'w')
Prob = file("Proj_Prob.txt","w")
#Pdb = atomwrite.pdbfile("anim.pdb",L, AtomNames = Names, Compressed=True)
#Pdb = atomwrite.pdbfile("animphob.pdb",L, AtomNames = Names, Compressed=True)
#Pdb = atomwrite.pdbfile("animcharge.pdb",L, AtomNames = Names, Compressed=True)
Pdb = atomwrite.pdbfile("animall.pdb",L, AtomNames = Names, Compressed=True)
rmin=5.0 #minimum contact distance
sigma = 2.5 #average length of amino acid not including R
rlj = 2.5*sigma  #cutoff for LJ potential
rw = 2**(1./6)*sigma #cutoff for WCA potential
rc = 26.5 #cutoff for electrostatic interactions
dt=0.001
T=1.0
EFracTolLS=1.e-8
EFracTolCG=1.e-10
Equilsteps=200  #time til equilibration (2e5 timesteps)
PdbFreq = 100
RescaleFreq=100
DistFreq=200
Trial = 0
while Trial<3:
    Pos=InitPositions(N,L)
    Vel=InitVelocities(N,T)
    Accel=InitAccel(Pos)
    PEnergy,Pos=ConjugateGradient(Pos,dt,EFracTolLS,EFracTolCG)
    P=np.zeros(shape=(N,N)) #matrix to keep track of whether within minimum distance
    Pnorm = [] #normalized probability
    Trial+=1
    t=0
    obs=0
    i=0 
    j=0  
    l=0 
    while t<1000:
        Pos,Vel,Accel,KEnergy,PEnergy=projlib.vvintegrate(Pos, Vel, Accel, M, L, AA, rc, rlj, rw, sigma, dt)
        if t<Equilsteps:
            if (i % RescaleFreq)==0:
                Vel=RescaleVelocities(Vel,T)
                i=0

        if t>Equilsteps:
            if Trial ==1 and (l%PdbFreq==0) and t<500:
                l=0        
                Pdb.write(Pos)
        
            if (j % DistFreq==0):
                if UseVisual:
                    if atomvis.Initialized:
                        #update the positions
                        atomvis.Update(Pos)
                    else:
                    #initialize the visualization window
                        atomvis.Init(Pos)
                obs+=1
            #PErun.append(PEnergy)
            #print PEnergy+KEnergy
                g.write('%.5f\t%.5f\n'%(PEnergy+KEnergy,t))
                f.write('%.5f\t%.5f\n'%(PEnergy,t))
                j=0
                Dist = projlib.distance(Pos,L)
                a=0
                while a<M:
                    b=0
                    while b<M:
                        if (Dist[a,b]<rmin):
                            if ((b<a-1) or (b>a+1)):
                            #print Dist[a,b]
                                P[a,b]+=1
                                #print P
                        b+=1
                    a+=1
        t+=dt
        i+=1
        j+=1
        l+=1

    Pnorm=np.array(P)/obs
    Prob.write('\n')
    for x in Pnorm: 
        Prob.write('%.5f\t%.5f\t%.5f\t%.5f\t%.5f\t%.5f\t%.5f\t%.5f\t%.5f\t%.5f\t%.5f\t%.5f\t%.5f\t%.5f\t%.5f\n'%(x[0],x[1],x[2],x[3],x[4],x[5],x[6],x[7],x[8],x[9],x[10],x[11],x[12],x[13],x[14])) 
    


#    print 'hello'
    
#print Pprint Pnorm
f.close()
g.close()
Pdb.close()
Prob.close()